"""Utils."""

import enum

from .errors import *

__all__ = [
    "PADDING_MODE",
    "DataPadder",
]


def ROL32(X: int, count: int):
    """Rotate left for 32 bit number."""

    count &= 0x1f
    return ((X << count) | (X >> (32 - count))) & 0xffffffff


def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, "big")


def int_to_bytes(i: int) -> bytes:
    """Convert integer to minimum number of bytes required to store its value."""

    return i.to_bytes((i.bit_length() + 7) >> 3, "big")


class PADDING_MODE(enum.Enum):
    """Block cipher data padding mode.

    Attributes:
        NONE: No padding.
        ZEROS: Append 0x00 bytes until the block size is reached.
        PKCS7: Append N bytes of value N, where N is the number of padding bytes.
        ISO7816_4: Append 0x80, followed by 0x00 bytes until the block size is reached.
        ANSI_X923: Append 0x00 bytes, with the last byte set to the number of padding bytes.
    """

    NONE = enum.auto()
    ZEROS = enum.auto()
    PKCS7 = enum.auto()
    ISO7816_4 = enum.auto()
    ANSI_X923 = enum.auto()


class DataPadder:
    """Data padder used to pad data to a multiple of block size."""

    def __init__(self, block_length: int = 1, method: PADDING_MODE = PADDING_MODE.NONE) -> None:
        """Data padder used to pad data to a multiple of block size.

        Args:
            block_length: Block size in bytes.
            method: Padding method.
        """

        if block_length <= 0:
            raise InvalidArgumentError("block_length must be positive")
        self._block_length = block_length
        self._method = method

    def pad(self, data: bytes) -> bytes:
        """Pad data to a multiple of block size.

        Args:
            data: Input data to be padded.

        Returns:
            bytes: Padded data.
        """

        if self._method is PADDING_MODE.NONE:
            return bytes(data)

        data = bytearray(data)

        rem = len(data) % self._block_length
        pad_len = self._block_length if rem == 0 else self._block_length - rem

        if self._method is PADDING_MODE.ZEROS:
            data.extend([0x00] * pad_len)
        elif self._method is PADDING_MODE.PKCS7:
            data.extend([pad_len] * pad_len)
        elif self._method is PADDING_MODE.ISO7816_4:
            data.append(0x80)
            data.extend([0x00] * (pad_len - 1))
        elif self._method is PADDING_MODE.ANSI_X923:
            data.extend([0x00] * (pad_len - 1))
            data.append(pad_len)
        else:
            raise NotImplementedError(f"Unsupported padding method: {self._method}")
        return bytes(data)

    def unpad(self, data: bytes) -> bytes:
        """Remove padding from data.

        Args:
            data: Padded data.

        Returns:
            bytes: Unpadded data.

        Raises:
            IncorrectLengthError: Invalid padding length.
            CheckFailedError: Invalid padding data.
        """

        if self._method is PADDING_MODE.NONE:
            return bytes(data)

        if not data or len(data) % self._block_length != 0:
            raise IncorrectLengthError("data", f"a multiple of {self._block_length} bytes", f"{len(data)} bytes")

        if self._method is PADDING_MODE.ZEROS:
            data = data.rstrip(b"\x00")
        elif self._method is PADDING_MODE.PKCS7:
            pad_len = data[-1]
            if pad_len == 0 or pad_len > self._block_length:
                raise IncorrectLengthError("PKCS7 last padding byte value", f"positive integer value smaller than {self._block_length}", f"{pad_len}")
            if data[-pad_len:] != bytes([pad_len]) * pad_len:
                raise CheckFailedError("Invalid PKCS7 padding bytes")
            data = data[:-pad_len]
        elif self._method is PADDING_MODE.ISO7816_4:
            # Must end with 0x80 followed by zero or more 0x00
            i = data.rfind(b"\x80")
            if i == -1 or any(b != 0x00 for b in data[i + 1:]):
                raise CheckFailedError("Invalid ISO7816-4 padding")
            data = data[:i]
        elif self._method is PADDING_MODE.ANSI_X923:
            pad_len = data[-1]
            if pad_len == 0 or pad_len > self._block_length:
                raise IncorrectLengthError("ANSI X9.23 last padding byte value", f"positive integer value smaller than {self._block_length}", f"{pad_len}")
            if any(b != 0x00 for b in data[-pad_len:-1]):
                raise CheckFailedError("Invalid ANSI X9.23 padding bytes")
            data = data[:-pad_len]
        else:
            raise NotImplementedError(f"Unsupported padding method: {self._method}")

        return bytes(data)



"""Block cipher with working mode."""

import enum

from .base import BlockCipher
from .errors import *

__all__ = [
    "BlockCipherMode",
    "BlockCipherModeECB",
    "BlockCipherModeCBC",
    "BlockCipherModeCFB",
    "BlockCipherModeOFB",
]


class BlockCipherMode:
    """Base class of block cipher with working mode."""

    def __init__(self, bc: BlockCipher) -> None:
        """Base class of block cipher with working mode.

        Args:
            bc (BlockCipher): Block cipher object used in en/decryption.
        """

        self._bc = bc
        self._block_len = bc.block_length()

    def reset(self):
        """Reset internal states."""

        raise NotImplementedError

    def encrypt(self, plain: bytes) -> bytes:
        """Encrypt."""

        raise NotImplementedError

    def decrypt(self, cipher: bytes) -> bytes:
        """Decrypt."""

        raise NotImplementedError


class BlockCipherModeECB(BlockCipherMode):
    """ECB block cipher working mode."""

    def __init__(self, bc: BlockCipher) -> None:
        """ECB block cipher working mode.

        Args:
            bc (BlockCipher): Block cipher object used in en/decryption.
        """

        super().__init__(bc)

    def reset(self):
        """Reset internal states."""

        return

    def encrypt(self, plain: bytes) -> bytes:
        """Encrypt.

        Args:
            plain (bytes): Data to be encrypted, must be padded as a multiple of block length.

        Returns:
            cipher: Encrypted plain data.

        Raises:
            IncorrectLengthError: Incorrect plain length.
        """

        plain_len = len(plain)

        if plain_len % self._block_len != 0:
            raise IncorrectLengthError("plain", f"a multiple of {self._block_len} bytes", f"{plain_len} bytes")

        cipher_blocks = []
        for i in range(0, plain_len, self._block_len):
            plain_block = plain[i:i+self._block_len]
            cipher_block = self._bc.encrypt(plain_block)
            cipher_blocks.append(cipher_block)

        return b"".join(cipher_blocks)

    def decrypt(self, cipher: bytes) -> bytes:
        """Decrypt.

        Args:
            cipher (bytes): Data to be decrypted.

        Returns:
            plain: Decrypted data, you may need to remove padding manually.

        Raises:
            IncorrectLengthError: Incorrect cipher length.
        """

        cipher_len = len(cipher)

        if cipher_len % self._block_len != 0:
            raise IncorrectLengthError("cipher", f"a multiple of {self._block_len} bytes", f"{cipher_len} bytes")

        plain_blocks = []
        for i in range(0, cipher_len, self._block_len):
            cipher_block = cipher[i:i+self._block_len]
            plain_block = self._bc.decrypt(cipher_block)
            plain_blocks.append(plain_block)

        return b"".join(plain_blocks)


class BlockCipherModeCBC(BlockCipherMode):
    """CBC block cipher working mode."""

    def __init__(self, bc: BlockCipher, iv: bytes) -> None:
        """CBC block cipher working mode.

        Args:
            bc (BlockCipher): Block cipher object used in en/decryption.
            iv (bytes): Initial vector, must be the same length as block length.

        Raises:
            IncorrectLengthError: Incorrect iv length.
        """

        super().__init__(bc)

        if len(iv) != bc.block_length():
            raise IncorrectLengthError("IV", f"{bc.block_length()} bytes", f"{len(iv)} bytes")

        self._iv = bytes(iv)
        self._last_cipher_block = self._iv

    def reset(self):
        """Reset internal states."""

        self._last_cipher_block = self._iv

    def encrypt(self, plain: bytes) -> bytes:
        """Encrypt.

        Args:
            plain (bytes): Data to be encrypted, must be padded as a multiple of block length.

        Returns:
            cipher: Encrypted plain data.

        Raises:
            IncorrectLengthError: Incorrect plain length.
        """

        plain_len = len(plain)

        if plain_len % self._block_len != 0:
            raise IncorrectLengthError("plain", f"a multiple of {self._block_len} bytes", f"{plain_len} bytes")

        cipher_blocks = []
        for i in range(0, plain_len, self._block_len):
            plain_block = plain[i:i+self._block_len]
            xor_block = bytes(b1 ^ b2 for b1, b2 in zip(self._last_cipher_block, plain_block))
            cipher_block = self._bc.encrypt(xor_block)
            cipher_blocks.append(cipher_block)
            self._last_cipher_block = cipher_block

        return b"".join(cipher_blocks)

    def decrypt(self, cipher: bytes) -> bytes:
        """Decrypt.

        Args:
            cipher (bytes): Data to be decrypted.

        Returns:
            plain: Decrypted data, you may need to remove padding manually.

        Raises:
            IncorrectLengthError: Incorrect cipher length.
        """

        cipher_len = len(cipher)

        if cipher_len % self._block_len != 0:
            raise IncorrectLengthError("cipher", f"a multiple of {self._block_len} bytes", f"{cipher_len} bytes")

        plain_blocks = []
        for i in range(0, cipher_len, self._block_len):
            cipher_block = cipher[i:i+self._block_len]
            xor_block = self._bc.decrypt(cipher_block)
            plain_block = bytes(b1 ^ b2 for b1, b2 in zip(self._last_cipher_block, xor_block))
            plain_blocks.append(plain_block)
            self._last_cipher_block = bytes(cipher_block)

        return b"".join(plain_blocks)


class BlockCipherModeCFB(BlockCipherMode):
    """CFB block cipher working mode."""

    def __init__(self, bc: BlockCipher, iv: bytes, segment_length: int = 1) -> None:
        """CFB block cipher working mode.

        Args:
            bc (BlockCipher): Block cipher object used in en/decryption.
            iv (bytes): Initial vector, must be the same length as block length.
            segment_length (int): Segment length of data stream in bytes,
                must be little or equal than block length, default to 1 (per-byte).

        Raises:
            IncorrectLengthError: Incorrect iv length.
            InvalidArgumentError: Invalid segment length.
        """

        super().__init__(bc)

        if len(iv) != bc.block_length():
            raise IncorrectLengthError("iv", f"{bc.block_length()} bytes", f"{len(iv)} bytes")

        if segment_length > bc.block_length():
            raise InvalidArgumentError(f"segment length ({segment_length}) can't be greater than block length ({bc.block_length()}).")

        self._iv = bytes(iv)
        self._segment_len = segment_length
        self._shift_register = self._iv
        self._key_stream = self._bc.encrypt(self._shift_register)[:self._segment_len]
        self._init_key_stream = self._key_stream

    def reset(self):
        """Reset internal states."""

        self._shift_register = self._iv
        self._key_stream = self._init_key_stream

    def encrypt(self, plain: bytes) -> bytes:
        """Encrypt.

        Args:
            plain (bytes): Data to be encrypted, no need to be padded.

        Returns:
            cipher: Encrypted plain data.
        """

        if len(plain) <= 0:
            return b""

        cipher_head = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, plain))
        head_len = len(cipher_head)
        self._shift_register = self._shift_register[head_len:] + cipher_head
        self._key_stream = self._key_stream[head_len:]

        if len(self._key_stream) > 0:
            return cipher_head

        self._key_stream = self._bc.encrypt(self._shift_register)[:self._segment_len]

        plain = plain[head_len:]
        plain_len = len(plain)
        if plain_len <= 0:
            return cipher_head

        tail_len = plain_len % self._segment_len
        plain_segments_len = plain_len - tail_len

        cipher_segments = [cipher_head]
        for i in range(0, plain_segments_len, self._segment_len):
            plain_segment = plain[i:i+self._segment_len]
            cipher_segment = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, plain_segment))
            cipher_segments.append(cipher_segment)
            self._shift_register = self._shift_register[self._segment_len:] + cipher_segment
            self._key_stream = self._bc.encrypt(self._shift_register)[:self._segment_len]

        if tail_len > 0:
            plain_tail = plain[plain_segments_len:]
            cipher_tail = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, plain_tail))
            cipher_segments.append(cipher_tail)
            self._shift_register = self._shift_register[tail_len:] + cipher_tail
            self._key_stream = self._key_stream[tail_len:]

        return b"".join(cipher_segments)

    def decrypt(self, cipher: bytes) -> bytes:
        """Decrypt.

        Args:
            cipher (bytes): Data to be decrypted.

        Returns:
            plain: Decrypted data, you may need to remove padding manually.
        """

        if len(cipher) <= 0:
            return b""

        plain_head = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, cipher))
        head_len = len(plain_head)
        self._shift_register = self._shift_register[head_len:] + cipher[:head_len]
        self._key_stream = self._key_stream[head_len:]

        if len(self._key_stream) > 0:
            return plain_head

        self._key_stream = self._bc.encrypt(self._shift_register)[:self._segment_len]

        cipher = cipher[head_len:]
        cipher_len = len(cipher)
        if cipher_len <= 0:
            return plain_head

        tail_len = cipher_len % self._segment_len
        cipher_segments_len = cipher_len - tail_len

        plain_segments = [plain_head]
        for i in range(0, cipher_segments_len, self._segment_len):
            cipher_segment = cipher[i:i+self._segment_len]
            plain_segment = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, cipher_segment))
            plain_segments.append(plain_segment)
            self._shift_register = self._shift_register[self._segment_len:] + cipher_segment
            self._key_stream = self._bc.encrypt(self._shift_register)[:self._segment_len]

        if tail_len > 0:
            cipher_tail = cipher[cipher_segments_len:]
            plain_tail = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, cipher_tail))
            plain_segments.append(plain_tail)
            self._shift_register = self._shift_register[tail_len:] + cipher_tail
            self._key_stream = self._key_stream[tail_len:]

        return b"".join(plain_segments)


class BlockCipherModeOFB(BlockCipherMode):
    """OFB block cipher working mode."""

    def __init__(self, bc: BlockCipher, iv: bytes) -> None:
        """OFB block cipher working mode.

        Args:
            bc (BlockCipher): Block cipher object used in en/decryption.
            iv (bytes): Initial vector, must be the same length as block length.

        Raises:
            IncorrectLengthError: Incorrect iv length.
        """

        super().__init__(bc)

        if len(iv) != bc.block_length():
            raise IncorrectLengthError("iv", f"{bc.block_length()} bytes", f"{len(iv)} bytes")

        self._key_stream = self._bc.encrypt(iv)
        self._next_key_block_in = self._key_stream
        self._init_key_stream = self._key_stream

    def reset(self):
        """Reset internal states."""

        self._key_stream = self._init_key_stream
        self._next_key_block_in = self._key_stream

    def encrypt(self, plain: bytes) -> bytes:
        """Encrypt.

        Args:
            plain (bytes): Data to be encrypted, no need to be padded.

        Returns:
            cipher: Encrypted plain data.
        """

        if len(plain) <= 0:
            return b""

        cipher_head = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, plain))
        head_len = len(cipher_head)
        self._key_stream = self._key_stream[head_len:]

        if len(self._key_stream) > 0:
            return cipher_head

        self._key_stream = self._bc.encrypt(self._next_key_block_in)
        self._next_key_block_in = self._key_stream

        plain = plain[head_len:]
        plain_len = len(plain)
        if plain_len <= 0:
            return cipher_head

        tail_len = plain_len % self._block_len
        plain_blocks_len = plain_len - tail_len

        cipher_blocks = [cipher_head]
        for i in range(0, plain_blocks_len, self._block_len):
            plain_block = plain[i:i+self._block_len]
            cipher_block = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, plain_block))
            cipher_blocks.append(cipher_block)
            self._key_stream = self._bc.encrypt(self._next_key_block_in)
            self._next_key_block_in = self._key_stream

        if tail_len > 0:
            plain_tail = plain[plain_blocks_len:]
            cipher_tail = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, plain_tail))
            cipher_blocks.append(cipher_tail)
            self._key_stream = self._key_stream[tail_len:]

        return b"".join(cipher_blocks)

    def decrypt(self, cipher: bytes) -> bytes:
        """Decrypt.

        Args:
            cipher (bytes): Data to be decrypted.

        Returns:
            plain: Decrypted data, you may need to remove padding manually.
        """

        if len(cipher) <= 0:
            return b""

        plain_head = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, cipher))
        head_len = len(plain_head)
        self._key_stream = self._key_stream[head_len:]

        if len(self._key_stream) > 0:
            return plain_head

        self._key_stream = self._bc.encrypt(self._next_key_block_in)
        self._next_key_block_in = self._key_stream

        cipher = cipher[head_len:]
        cipher_len = len(cipher)
        if cipher_len <= 0:
            return plain_head

        tail_len = cipher_len % self._block_len
        cipher_blocks_len = cipher_len - tail_len

        plain_blocks = [plain_head]
        for i in range(0, cipher_blocks_len, self._block_len):
            cipher_block = cipher[i:i+self._block_len]
            plain_block = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, cipher_block))
            plain_blocks.append(plain_block)
            self._key_stream = self._bc.encrypt(self._next_key_block_in)
            self._next_key_block_in = self._key_stream

        if tail_len > 0:
            cipher_tail = cipher[cipher_blocks_len:]
            plain_tail = bytes(b1 ^ b2 for b1, b2 in zip(self._key_stream, cipher_tail))
            plain_blocks.append(plain_tail)
            self._key_stream = self._key_stream[tail_len:]

        return b"".join(plain_blocks)



"""SM3 Algorithm Implementation Module."""

from typing import List

from .base import Hash
from .errors import *
from .utils import ROL32

__all__ = ["SM3"]

_ROL_T_TABLE = [
    0x79cc4519, 0xf3988a32, 0xe7311465, 0xce6228cb, 0x9cc45197, 0x3988a32f, 0x7311465e, 0xe6228cbc,
    0xcc451979, 0x988a32f3, 0x311465e7, 0x6228cbce, 0xc451979c, 0x88a32f39, 0x11465e73, 0x228cbce6,
    0x9d8a7a87, 0x3b14f50f, 0x7629ea1e, 0xec53d43c, 0xd8a7a879, 0xb14f50f3, 0x629ea1e7, 0xc53d43ce,
    0x8a7a879d, 0x14f50f3b, 0x29ea1e76, 0x53d43cec, 0xa7a879d8, 0x4f50f3b1, 0x9ea1e762, 0x3d43cec5,
    0x7a879d8a, 0xf50f3b14, 0xea1e7629, 0xd43cec53, 0xa879d8a7, 0x50f3b14f, 0xa1e7629e, 0x43cec53d,
    0x879d8a7a, 0x0f3b14f5, 0x1e7629ea, 0x3cec53d4, 0x79d8a7a8, 0xf3b14f50, 0xe7629ea1, 0xcec53d43,
    0x9d8a7a87, 0x3b14f50f, 0x7629ea1e, 0xec53d43c, 0xd8a7a879, 0xb14f50f3, 0x629ea1e7, 0xc53d43ce,
    0x8a7a879d, 0x14f50f3b, 0x29ea1e76, 0x53d43cec, 0xa7a879d8, 0x4f50f3b1, 0x9ea1e762, 0x3d43cec5,
]


def _precomp_rol_table():
    return [ROL32(_T(i), i) for i in range(64)]


def _T(i):
    return 0x79cc4519 if i <= 15 else 0x7a879d8a


def _FF(i, X, Y, Z):
    return (X ^ Y ^ Z) if i <= 15 else ((X & Y) | (X & Z) | (Y & Z))


def _GG(i, X, Y, Z):
    return (X ^ Y ^ Z) if i <= 15 else ((X & Y) | (~X & Z))


def _P0(X):
    return X ^ ROL32(X, 9) ^ ROL32(X, 17)


def _P1(X):
    return X ^ ROL32(X, 15) ^ ROL32(X, 23)


def _expand(B: bytes, W1: List[int], W2: List[int]):
    """Expand message block."""

    for i in range(16):
        W1[i] = int.from_bytes(B[i * 4:i * 4 + 4], "big")
    for i in range(16, 68):
        W1[i] = _P1(W1[i - 16] ^ W1[i - 9] ^ ROL32(W1[i - 3], 15)) ^ ROL32(W1[i - 13], 7) ^ W1[i - 6]
    for i in range(64):
        W2[i] = W1[i] ^ W1[i + 4]


def _compress(W1: List[int], W2: List[int], V: List[int]):
    """Compress words."""

    A, B, C, D, E, F, G, H = V

    for i in range(64):
        SS1 = ROL32((ROL32(A, 12) + E + _ROL_T_TABLE[i]) & 0xffffffff, 7)
        SS2 = SS1 ^ ROL32(A, 12)
        TT1 = (_FF(i, A, B, C) + D + SS2 + W2[i]) & 0xffffffff
        TT2 = (_GG(i, E, F, G) + H + SS1 + W1[i]) & 0xffffffff
        D = C
        C = ROL32(B, 9)
        B = A
        A = TT1
        H = G
        G = ROL32(F, 19)
        F = E
        E = _P0(TT2)

    V[0] ^= A
    V[1] ^= B
    V[2] ^= C
    V[3] ^= D
    V[4] ^= E
    V[5] ^= F
    V[6] ^= G
    V[7] ^= H


class SM3(Hash):
    """SM3 Algorithm."""

    @classmethod
    def max_msg_length(cls) -> int:
        """Get maximum message length in bytes."""

        return 0x1fffffffffffffff  # (1 << 64 - 1) >> 3

    @classmethod
    def hash_length(cls) -> int:
        """Get output hash value length in bytes."""

        return 32

    def __init__(self) -> None:
        """SM3 Algorithm."""

        self._value: List[int] = [0x7380166f, 0x4914b2b9, 0x172442d7, 0xda8a0600, 0xa96f30bc, 0x163138aa, 0xe38dee4d, 0xb0fb0e4e]
        self._msg_len: int = 0
        self._msg_block_buffer: bytearray = bytearray()

        self._words_buffer1: List[int] = [0] * 68
        self._words_buffer2: List[int] = [0] * 64

    def reset(self):
        """Reset to initial state."""

        self.__init__()

    def update(self, data: bytes) -> None:
        """Update internal state.

        Args:
            data: Data stream to be updated.

        Raises:
            DataOverflowError: Message too long.
        """

        if self._msg_len + len(data) > self.max_msg_length():
            raise DataOverflowError("Message", f"0x{self.max_msg_length():x} bytes")

        B = self._msg_block_buffer
        W1 = self._words_buffer1
        W2 = self._words_buffer2
        V = self._value

        b_len = len(B)
        d_len = len(data)
        if b_len + d_len >= 64:
            # process last short block
            begin = 64 - b_len
            B.extend(data[:begin])
            _expand(B, W1, W2)
            _compress(W1, W2, V)
            B.clear()

            pos = begin
            while pos + 63 < d_len:
                _expand(data[pos:pos+64], W1, W2)
                _compress(W1, W2, V)
                pos += 64

            B.extend(data[pos:])
        else:
            B.extend(data)

        self._msg_len += d_len

    def value(self) -> bytes:
        """Get current hash value.

        Returns:
            bytes: Current hash value.

        Note:
            Each invocation of the method will dynamically compute the hash value that should be output,
                so it is advisable to retain the resulting hash value after the method call.
        """

        B = self._msg_block_buffer.copy()
        W1 = self._words_buffer1
        W2 = self._words_buffer2
        V = self._value.copy()

        b_len = len(B)

        B.append(0x80)

        if b_len < 56:
            for _ in range(b_len + 1, 56):
                B.append(0x00)
        else:
            for _ in range(b_len + 1, 64):
                B.append(0x00)
            _expand(B, W1, W2)
            _compress(W1, W2, V)
            B = bytearray(56)

        B.extend((self._msg_len << 3).to_bytes(8, "big"))

        _expand(B, W1, W2)
        _compress(W1, W2, V)

        value = bytearray()
        for w in V:
            value.extend(w.to_bytes(4, "big"))
        return bytes(value)



"""Elliptic curve operations.

This module provides fundamental operations related to elliptic curves,
    including but not limited to those specified in the national cryptographic standard documents.
"""

from typing import Tuple, Union

from . import primefield as Fp
from .errors import *

__all__ = [
    "EcPoint",
    "EcPoint2",
    "EcPoint4",
    "EcPoint12",
    "EcPointEx",
    "EllipticCurve",
    "ECDLP",
    "SM9BNBP",
]

EcPoint = Tuple[int, int]
EcPoint2 = Tuple[Fp.Fp2Ele, Fp.Fp2Ele]
EcPoint4 = Tuple[Fp.Fp4Ele, Fp.Fp4Ele]
EcPoint12 = Tuple[Fp.Fp12Ele, Fp.Fp12Ele]
EcPointEx = Tuple[Fp.FpExEle, Fp.FpExEle]


class EllipticCurve:
    """Elliptic Curve.

    Attributes:
        INF: Infinite point.
        a (FpExEle): Parameter a of curve.
        b (FpExEle): Parameter b of curve.
    """

    INF = (float("inf"), float("inf"))

    def __init__(self, fp: Fp.PrimeFieldBase, a: Fp.FpExEle, b: Fp.FpExEle) -> None:
        """Elliptic curve.

        Args:
            fp: Prime field operations used in ellitic curve, must be derived from class `PrimeFieldBase`. 
            a: Parameter a of curve.
            b: Parameter b of curve.

        Note:
            When instantiating the class, the `FpExEle` or `EcPointEx` type in the methods will correspond to the type of `fp`.
                Here, a generic placeholder are used to denote the data type.
        """

        self.a = a
        self.b = b
        self._fp = fp

    def get_y_sqr(self, x: Fp.FpExEle) -> Fp.FpExEle:
        """Get the square of y for the specified x."""

        fp = self._fp
        return fp.add(fp.pow(x, 3), fp.add(fp.mul(self.a, x), self.b))

    def get_y(self, x: Fp.FpExEle) -> Union[Fp.FpExEle, None]:
        """Get one of valid y for given x, `None` means no solution."""

        return self._fp.sqrt(self.get_y_sqr(x))

    def isvalid(self, P: EcPointEx) -> bool:
        """Whether the point is on curve."""

        x, y = P
        return self._fp.mul(y, y) == self.get_y_sqr(x)

    def neg(self, P: EcPointEx) -> EcPointEx:
        """Get negative point."""

        x, y = P
        return (x, self._fp.neg(y))

    def add(self, P1: EcPointEx, P2: EcPointEx) -> EcPointEx:
        """Add two points."""

        fp = self._fp

        if P1 == self.INF:
            return P2
        if P2 == self.INF:
            return P1

        x1, y1 = P1
        x2, y2 = P2

        if x1 == x2:
            if fp.isoppo(y1, y2):
                return self.INF
            elif y1 == y2:
                _t1 = fp.add(self.a, fp.smul(3, fp.mul(x1, x1)))
                _t2 = fp.inv(fp.smul(2, y1))
                lam = fp.mul(_t1, _t2)
            else:
                raise UnknownError(f"y1 and y2 is neither equal nor opposite.")
        else:
            lam = fp.mul(fp.sub(y2, y1), fp.inv(fp.sub(x2, x1)))

        x3 = fp.sub(fp.mul(lam, lam), fp.add(x1, x2))
        y3 = fp.sub(fp.mul(lam, fp.sub(x1, x3)), y1)
        return x3, y3

    def sub(self, P1: EcPointEx, P2: EcPointEx) -> EcPointEx:
        """Substract two points."""

        return self.add(P1, self.neg(P2))

    def mul(self, k: int, P: EcPointEx) -> EcPointEx:
        """Scalar multiplication of point by k."""

        Q = P
        for i in f"{k:b}"[1:]:
            Q = self.add(Q, Q)
            if i == "1":
                Q = self.add(Q, P)
        return Q


class ECDLP:
    """Elliptic Curve Discrete Logarithm Problem.

    Attributes:
        fp (PrimeField): `PrimeField` used in ECDLP.
        ec (EllipticCurve): `EllipticCurve` used in ECDLP.
        G (EcPoint): Base point.
        fpn (PrimeField): `PrimeField` operations for the order of base point.
        h (int): Cofactor of base point.
    """

    def __init__(self, p: int, a: int, b: int, G: EcPoint, n: int, h: int = 1) -> None:
        """Elliptic Curve Discrete Logarithm Problem.

        Args:
            p: Parameter p of curve.
            a: Parameter a of curve.
            b: Parameter b of curve.
            G: Base point.
            n: Order of base point.
            h: Cofactor of `G`, default to `1`.
        """

        self.fp = Fp.PrimeField(p)
        self.ec = EllipticCurve(self.fp, a, b)
        self.G = G
        self.fpn = Fp.PrimeField(n)
        self.h = h

    def kG(self, k: int) -> EcPoint:
        """Scalar multiplication of G by k."""

        return self.ec.mul(k, self.G)


class SM9BNBP:
    """SM9 Bilinear Pairing on Barreto-Naehrig (BN) Elliptic Curve.

    For performance reasons,
        this implementation focuses solely on operations on the BN curve in the context of SM9.
        Consequently, certain intermediate values can be precomputed to enhance computational efficiency.

    The fundamental parameters used in SM9 are as follows:

    ```python
    t = 0x600000000058F98A
    p = 0xB640000002A3A6F1D603AB4FF58EC74521F2934B1A7AEEDBE56F9B27E351457D
    n = 0xB640000002A3A6F1D603AB4FF58EC74449F2934B18EA8BEEE56EE19CD69ECF25
    b = 0x05
    beta_b = (5, 0)
    ```

    There are some more precomputed intermediate values.
        Refer to the source code for detailed information.

    Attributes:
        t (int): Parameter t of SM9.
        fp12 (PrimeField12): `PrimeField12` operations used in SM9.
        fp2 (PrimeField2): `PrimeField2` operations used in SM9.
        fp1 (PrimeField): `PrimeField` operations used in SM9.
        fpn (PrimeField): `PrimeField` operations for the order of base point.
        ec1 (EllipticCurve): `EllipticCurve` on Fp operations used in SM9.
        ec2 (EllipticCurve): `EllipticCurve` on Fp2 operations used in SM9.
        G1 (EcPoint): Base point of group 1.
        G2 (EcPoint2): Base point of group 2.
    """

    def __init__(self, G1: EcPoint, G2: EcPoint2) -> None:
        """SM9 Bilinear Pairing on Barreto-Naehrig (BN) Elliptic Curve.

        Args:
            G1: Base point of group 1.
            G2: Base point of group 2.
        """

        # SM9 parameters

        t = 0x600000000058F98A
        p = 0xB640000002A3A6F1D603AB4FF58EC74521F2934B1A7AEEDBE56F9B27E351457D  # 36 * t**4 + 36 * t**3 + 24 * t**2 + 6 * t + 1
        n = 0xB640000002A3A6F1D603AB4FF58EC74449F2934B18EA8BEEE56EE19CD69ECF25  # 36 * t**4 + 36 * t**3 + 18 * t**2 + 6 * t + 1
        b = 0x05
        beta_b = (5, 0)  # (1, 0) * (0, 5)

        self.t = t

        self.fp12 = Fp.PrimeField12(p)
        self.fp2 = self.fp12.fp4.fp2
        self.fp1 = self.fp2.fp
        self.fpn = Fp.PrimeField(n)

        self.ec1 = EllipticCurve(self.fp1, 0, b)
        self.ec2 = EllipticCurve(self.fp2, (0, 0), beta_b)

        self.G1 = G1
        self.G2 = G2

        # self._a = 0x2400000000215D93E  # 6 * t + 2
        self._e_a = "00100000000000000000000000000000000000010000101011101100100111110"  # f"{self._a:b}"[1:]

        # phi factors

        self._neg2 = 0xB640000002A3A6F1D603AB4FF58EC74521F2934B1A7AEEDBE56F9B27E351457B  # self.fp1.neg(2)
        self._inv_neg2 = 0x5B2000000151D378EB01D5A7FAC763A290F949A58D3D776DF2B7CD93F1A8A2BE  # self.fp1.inv(self._neg2)

        # Frobenius factors

        w5 = 0x2D40A38CF6983351711E5F99520347CC57D778A9F8FF4C8A4C949C7FA2A96686
        w4 = 0xF300000002A3A6F2780272354F8B78F4D5FC11967BE65333
        w3 = 0x6C648DE5DC0A3F2CF55ACC93EE0BAF159F9D411806DC5177F5B21FD3DA24D011
        w2 = 0xF300000002A3A6F2780272354F8B78F4D5FC11967BE65334
        w1 = 0x3F23EA58E5720BDB843C6CFA9C08674947C5C86E0DDD04EDA91D8354377B698B
        w0 = 0x1

        self._frob1_factor = (((p - w5, w5), (p - w2, w2)), ((p - w4, w4), (p - w1, w1)), ((p - w3, w3), (p - w0, w0)))
        self._frob2_factor = (((p - w4, p - w4), (w4, w4)), ((p - w2, p - w2), (w2, w2)), ((p - w0, p - w0), (w0, w0)))
        self._frob3_factor = (((p - w3, w3), (w0, p - w0)), ((p - w0, w0), (p - w3, w3)), ((w3, p - w3), (p - w0, w0)))
        self._frob6_factor = (((p - w0, p - w0), (w0, w0)), ((w0, w0), (p - w0, p - w0)), ((p - w0, p - w0), (w0, w0)))

    def kG1(self, k: int) -> EcPoint:
        """Scalar multiplication of G1 by k."""

        return self.ec1.mul(k, self.G1)

    def kG2(self, k: int) -> EcPoint2:
        """Scalar multiplication of G2 by k."""

        return self.ec2.mul(k, self.G2)

    def _g_fn(self, U: EcPoint12, V: EcPoint12, Q: EcPoint12) -> Fp.Fp12Ele:
        """g(U, V)(Q).

        U, V, Q are Fp12 points.
        """

        fp12 = self.fp12

        if U == EllipticCurve.INF or V == EllipticCurve.INF or Q == EllipticCurve.INF:
            return fp12.one()

        xU, yU = U
        xV, yV = V
        xQ, yQ = Q

        if xU == xV:
            if fp12.isoppo(yU, yV):
                return fp12.sub(xQ, xV), fp12.one()
            elif yU == yV:
                lam = fp12.mul(
                    fp12.smul(3, fp12.mul(xV, xV)),
                    fp12.inv(fp12.smul(2, yV))
                )
            else:
                raise UnknownError(f"y1 and y2 is neither equal nor opposite.")
        else:
            lam = fp12.mul(fp12.sub(yU, yV), fp12.inv(fp12.sub(xU, xV)))

        g = fp12.sub(fp12.mul(lam, fp12.sub(xQ, xV)), fp12.sub(yQ, yV))
        return g

    def _phi(self, P: EcPoint2) -> EcPoint12:
        """Get x, y in E (Fp12) from E' (Fp2), only implemented for beta=(1, 0)."""

        fp1 = self.fp1
        _i2 = self._inv_neg2

        x_, y_ = P

        x: Fp.Fp12Ele = (((0, 0), (0, 0)), ((fp1.mul(x_[1], _i2), x_[0]), (0, 0)), ((0, 0), (0, 0)))
        y: Fp.Fp12Ele = (((0, 0), (0, 0)), ((0, 0), (0, 0)), ((fp1.mul(y_[1], _i2), y_[0]), (0, 0)))

        return x, y

    def _phi_inv(self, P: EcPoint12) -> EcPoint2:
        """Inversion of `phi`."""

        fp1 = self.fp1
        _2 = self._neg2

        x_, y_ = P

        x: Fp.Fp2Ele = (x_[1][0][1], fp1.mul(x_[1][0][0], _2))
        y: Fp.Fp2Ele = (y_[2][0][1], fp1.mul(y_[2][0][0], _2))

        return x, y

    def _frob1(self, X: Fp.Fp12Ele) -> Fp.Fp12Ele:
        return self.fp12.pmul(X, self._frob1_factor)

    def _frob2(self, X: Fp.Fp12Ele) -> Fp.Fp12Ele:
        return self.fp12.pmul(X, self._frob2_factor)

    def _frob3(self, X: Fp.Fp12Ele) -> Fp.Fp12Ele:
        return self.fp12.pmul(X, self._frob3_factor)

    def _frob6(self, X: Fp.Fp12Ele) -> Fp.Fp12Ele:
        fp4 = self.fp12.fp4
        X2, X1, X0 = X
        return (fp4.conj(X2), fp4.neg(fp4.conj(X1)), fp4.conj(X0))

    def _finalexp(self, f: Fp.Fp12Ele) -> Fp.Fp12Ele:
        fp12 = self.fp12
        M = fp12.mul
        I = fp12.inv
        P = fp12.pow
        F1 = self._frob1
        F2 = self._frob2
        F3 = self._frob3
        F6 = self._frob6

        # easy part
        f = M(F6(f), I(f))
        f = M(F2(f), f)

        # hard part
        f_t = P(f, self.t)
        f_t2 = P(f_t, self.t)
        f_t3 = P(f_t2, self.t)

        f_p = F1(f)
        f_p2 = F2(f)
        f_p3 = F3(f)

        f_t_p = F1(f_t)
        f_t2_p = F1(f_t2)
        f_t3_p = F1(f_t3)
        f_t2_p2 = F2(f_t2)

        # y6, y5, y4, y3, y2, y1, y0
        #  -,  -,  -,  -,  +,  -,  +
        y6 = P(M(f_t3, f_t3_p), 36)
        y5 = P(f_t2, 30)
        y4 = P(M(f_t2_p, f_t), 18)
        y3 = P(f_t_p, 12)
        y2 = P(f_t2_p2, 6)
        y1 = M(f, f)
        y0 = M(f_p, M(f_p2, f_p3))

        f_num = M(y2, y0)
        f_den = M(y6, M(y5, M(y4, M(y3, y1))))

        f = M(f_num, I(f_den))
        return f

    def e(self, P: EcPoint, Q: EcPoint2) -> Fp.Fp12Ele:
        """R-ate bilinear pairing.

        Args:
            P: Element of group 1.
            Q: Element of group 2.

        Returns:
            Fp12Ele: Pairing value on Fp12.
        """

        fp12 = self.fp12
        ec2 = self.ec2
        phi = self._phi
        g_fn = self._g_fn

        _P = (fp12.extend(P[0]), fp12.extend(P[1]))  # P on E(Fp12)
        _Q = phi(Q)  # Q on E(Fp12)

        T = Q
        f = fp12.one()
        for i in self._e_a:
            _T = phi(T)  # T on E(Fp12)
            g = g_fn(_T, _T, _P)
            f = fp12.mul(fp12.mul(f, f), g)
            T = ec2.add(T, T)

            if i == "1":
                g = g_fn(phi(T), _Q, _P)
                f = fp12.mul(f, g)
                T = ec2.add(T, Q)

        _Q1 = (self._frob1(_Q[0]), self._frob1(_Q[1]))
        _Q2 = (self._frob2(_Q[0]), fp12.neg(self._frob2(_Q[1])))

        g = g_fn(phi(T), _Q1, _P)
        f = fp12.mul(f, g)

        T = ec2.add(T, self._phi_inv(_Q1))

        g = g_fn(phi(T), _Q2, _P)
        f = fp12.mul(f, g)

        f = self._finalexp(f)
        return f

    def eG1(self, Q: EcPoint2) -> Fp.Fp12Ele:
        """R-ate of G1 and Q."""

        return self.e(self.G1, Q)

    def eG2(self, P: EcPoint) -> Fp.Fp12Ele:
        """R-ate of P and G2."""

        return self.e(P, self.G2)



"""SM4 Algorithm Implementation Module."""

from typing import List

from .base import BlockCipher
from .errors import *
from .utils import ROL32

__all__ = ["SM4"]

_S_BOX = bytes([
    0xd6, 0x90, 0xe9, 0xfe, 0xcc, 0xe1, 0x3d, 0xb7, 0x16, 0xb6, 0x14, 0xc2, 0x28, 0xfb, 0x2c, 0x05,
    0x2b, 0x67, 0x9a, 0x76, 0x2a, 0xbe, 0x04, 0xc3, 0xaa, 0x44, 0x13, 0x26, 0x49, 0x86, 0x06, 0x99,
    0x9c, 0x42, 0x50, 0xf4, 0x91, 0xef, 0x98, 0x7a, 0x33, 0x54, 0x0b, 0x43, 0xed, 0xcf, 0xac, 0x62,
    0xe4, 0xb3, 0x1c, 0xa9, 0xc9, 0x08, 0xe8, 0x95, 0x80, 0xdf, 0x94, 0xfa, 0x75, 0x8f, 0x3f, 0xa6,
    0x47, 0x07, 0xa7, 0xfc, 0xf3, 0x73, 0x17, 0xba, 0x83, 0x59, 0x3c, 0x19, 0xe6, 0x85, 0x4f, 0xa8,
    0x68, 0x6b, 0x81, 0xb2, 0x71, 0x64, 0xda, 0x8b, 0xf8, 0xeb, 0x0f, 0x4b, 0x70, 0x56, 0x9d, 0x35,
    0x1e, 0x24, 0x0e, 0x5e, 0x63, 0x58, 0xd1, 0xa2, 0x25, 0x22, 0x7c, 0x3b, 0x01, 0x21, 0x78, 0x87,
    0xd4, 0x00, 0x46, 0x57, 0x9f, 0xd3, 0x27, 0x52, 0x4c, 0x36, 0x02, 0xe7, 0xa0, 0xc4, 0xc8, 0x9e,
    0xea, 0xbf, 0x8a, 0xd2, 0x40, 0xc7, 0x38, 0xb5, 0xa3, 0xf7, 0xf2, 0xce, 0xf9, 0x61, 0x15, 0xa1,
    0xe0, 0xae, 0x5d, 0xa4, 0x9b, 0x34, 0x1a, 0x55, 0xad, 0x93, 0x32, 0x30, 0xf5, 0x8c, 0xb1, 0xe3,
    0x1d, 0xf6, 0xe2, 0x2e, 0x82, 0x66, 0xca, 0x60, 0xc0, 0x29, 0x23, 0xab, 0x0d, 0x53, 0x4e, 0x6f,
    0xd5, 0xdb, 0x37, 0x45, 0xde, 0xfd, 0x8e, 0x2f, 0x03, 0xff, 0x6a, 0x72, 0x6d, 0x6c, 0x5b, 0x51,
    0x8d, 0x1b, 0xaf, 0x92, 0xbb, 0xdd, 0xbc, 0x7f, 0x11, 0xd9, 0x5c, 0x41, 0x1f, 0x10, 0x5a, 0xd8,
    0x0a, 0xc1, 0x31, 0x88, 0xa5, 0xcd, 0x7b, 0xbd, 0x2d, 0x74, 0xd0, 0x12, 0xb8, 0xe5, 0xb4, 0xb0,
    0x89, 0x69, 0x97, 0x4a, 0x0c, 0x96, 0x77, 0x7e, 0x65, 0xb9, 0xf1, 0x09, 0xc5, 0x6e, 0xc6, 0x84,
    0x18, 0xf0, 0x7d, 0xec, 0x3a, 0xdc, 0x4d, 0x20, 0x79, 0xee, 0x5f, 0x3e, 0xd7, 0xcb, 0x39, 0x48
])

_CK = [
    0x00070E15, 0x1C232A31, 0x383F464D, 0x545B6269, 0x70777E85, 0x8C939AA1, 0xA8AFB6BD, 0xC4CBD2D9,
    0xE0E7EEF5, 0xFC030A11, 0x181F262D, 0x343B4249, 0x50575E65, 0x6C737A81, 0x888F969D, 0xA4ABB2B9,
    0xC0C7CED5, 0xDCE3EAF1, 0xF8FF060D, 0x141B2229, 0x30373E45, 0x4C535A61, 0x686F767D, 0x848B9299,
    0xA0A7AEB5, 0xBCC3CAD1, 0xD8DFE6ED, 0xF4FB0209, 0x10171E25, 0x2C333A41, 0x484F565D, 0x646B7279
]


def _BS(X):
    return ((_S_BOX[(X >> 24) & 0xff] << 24) |
            (_S_BOX[(X >> 16) & 0xff] << 16) |
            (_S_BOX[(X >> 8) & 0xff] << 8) |
            (_S_BOX[X & 0xff]))


def _T0(X):
    X = _BS(X)
    return X ^ ROL32(X, 2) ^ ROL32(X, 10) ^ ROL32(X, 18) ^ ROL32(X, 24)


def _T1(X):
    X = _BS(X)
    return X ^ ROL32(X, 13) ^ ROL32(X, 23)


def _key_expand(key: bytes, rkey: List[int]):
    """Key expansion."""

    K0 = int.from_bytes(key[0:4], "big") ^ 0xa3b1bac6
    K1 = int.from_bytes(key[4:8], "big") ^ 0x56aa3350
    K2 = int.from_bytes(key[8:12], "big") ^ 0x677d9197
    K3 = int.from_bytes(key[12:16], "big") ^ 0xb27022dc

    for i in range(0, 32, 4):
        K0 = K0 ^ _T1(K1 ^ K2 ^ K3 ^ _CK[i])
        rkey[i] = K0
        K1 = K1 ^ _T1(K2 ^ K3 ^ K0 ^ _CK[i + 1])
        rkey[i + 1] = K1
        K2 = K2 ^ _T1(K3 ^ K0 ^ K1 ^ _CK[i + 2])
        rkey[i + 2] = K2
        K3 = K3 ^ _T1(K0 ^ K1 ^ K2 ^ _CK[i + 3])
        rkey[i + 3] = K3


class SM4(BlockCipher):
    """SM4 Algorithm."""

    @classmethod
    def key_length(cls) -> int:
        """Get key length in bytes."""

        return 16

    @classmethod
    def block_length(cls) -> int:
        """Get block length in bytes."""

        return 16

    def __init__(self, key: bytes) -> None:
        """SM4 Algorithm.

        Args:
            key: 16 bytes key.

        Raises:
            IncorrectLengthError: Incorrect key length.
        """

        if len(key) != self.key_length():
            raise IncorrectLengthError("Key", f"{self.key_length()} bytes", f"{len(key)} bytes")

        self._key: bytes = key
        self._rkey: List[int] = [0] * 32
        _key_expand(self._key, self._rkey)

        self._block_buffer = bytearray()

    def encrypt(self, block: bytes) -> bytes:
        """Encrypt.

        Args:
            block: Plain block to encrypt, must be 16 bytes.

        Returns:
            bytes: 16 bytes cipher block.

        Raises:
            IncorrectLengthError: Incorrect block length.
        """

        if len(block) != self.block_length():
            raise IncorrectLengthError("Block", f"{self.block_length()} bytes", f"{len(block)} bytes")

        RK = self._rkey

        X0 = int.from_bytes(block[0:4], "big")
        X1 = int.from_bytes(block[4:8], "big")
        X2 = int.from_bytes(block[8:12], "big")
        X3 = int.from_bytes(block[12:16], "big")

        for i in range(0, 32, 4):
            X0 = X0 ^ _T0(X1 ^ X2 ^ X3 ^ RK[i])
            X1 = X1 ^ _T0(X2 ^ X3 ^ X0 ^ RK[i + 1])
            X2 = X2 ^ _T0(X3 ^ X0 ^ X1 ^ RK[i + 2])
            X3 = X3 ^ _T0(X0 ^ X1 ^ X2 ^ RK[i + 3])

        BUFFER = self._block_buffer
        BUFFER.clear()
        BUFFER.extend(X3.to_bytes(4, "big"))
        BUFFER.extend(X2.to_bytes(4, "big"))
        BUFFER.extend(X1.to_bytes(4, "big"))
        BUFFER.extend(X0.to_bytes(4, "big"))
        return bytes(BUFFER)

    def decrypt(self, block: bytes) -> bytes:
        """Decrypt.

        Args:
            block: cipher block to decrypt, must be 16 bytes.

        Returns:
            bytes: 16 bytes plain block.

        Raises:
            IncorrectLengthError: Incorrect block length.
        """

        if len(block) != self.block_length():
            raise IncorrectLengthError("Block", f"{self.block_length()} bytes", f"{len(block)} bytes")

        RK = self._rkey

        X0 = int.from_bytes(block[0:4], "big")
        X1 = int.from_bytes(block[4:8], "big")
        X2 = int.from_bytes(block[8:12], "big")
        X3 = int.from_bytes(block[12:16], "big")

        for i in range(0, 32, 4):
            X0 = X0 ^ _T0(X1 ^ X2 ^ X3 ^ RK[31 - i])
            X1 = X1 ^ _T0(X2 ^ X3 ^ X0 ^ RK[30 - i])
            X2 = X2 ^ _T0(X3 ^ X0 ^ X1 ^ RK[29 - i])
            X3 = X3 ^ _T0(X0 ^ X1 ^ X2 ^ RK[28 - i])

        BUFFER = self._block_buffer
        BUFFER.clear()
        BUFFER.extend(X3.to_bytes(4, "big"))
        BUFFER.extend(X2.to_bytes(4, "big"))
        BUFFER.extend(X1.to_bytes(4, "big"))
        BUFFER.extend(X0.to_bytes(4, "big"))
        return bytes(BUFFER)



from . import errors
from .base import KEYXCHG_MODE, PC_MODE
from .sm2 import SM2
from .sm3 import SM3
from .sm4 import SM4
from .sm9 import SM9, SM9KGC
from .zuc import ZUC

__all__ = [
    "errors",
    "KEYXCHG_MODE",
    "PC_MODE",
    "SM2",
    "SM3",
    "SM4",
    "SM9",
    "SM9KGC",
    "ZUC",
]

__version__ = "1.1.1"
